# pylint: disable=wrong-import-position
"""
archinfo is a collection of classes that contain architecture-specific information.
It is useful for cross-architecture tools (such as pyvex).
"""

__version__ = "9.2.221"


from .arch import (
    Arch,
    ArchNotFound,
    Register,
    all_arches,
    arch_from_id,
    get_host_arch,
    register_arch,
    reverse_ends,
)
from .arch_aarch64 import ArchAArch64
from .arch_amd64 import ArchAMD64
from .arch_arm import ArchARM, ArchARMCortexM, ArchARMEL, ArchARMHF
from .arch_mips32 import ArchMIPS32
from .arch_mips64 import ArchMIPS64
from .arch_pcode import ArchPcode
from .arch_ppc32 import ArchPPC32
from .arch_ppc64 import ArchPPC64
from .arch_riscv64 import ArchRISCV64
from .arch_s390x import ArchS390X
from .arch_soot import ArchSoot
from .arch_x86 import ArchX86
from .archerror import ArchError
from .types import Endness, RegisterName, RegisterOffset, TmpVar

__all__ = [
    "Arch",
    "ArchAArch64",
    "ArchAMD64",
    "ArchARM",
    "ArchARMCortexM",
    "ArchARMEL",
    "ArchARMHF",
    "ArchError",
    "ArchMIPS32",
    "ArchMIPS64",
    "ArchNotFound",
    "ArchPPC32",
    "ArchPPC64",
    "ArchPcode",
    "ArchRISCV64",
    "ArchS390X",
    "ArchSoot",
    "ArchX86",
    "Endness",
    "Register",
    "RegisterName",
    "RegisterOffset",
    "TmpVar",
    "all_arches",
    "arch_from_id",
    "get_host_arch",
    "register_arch",
    "reverse_ends",
]



from archinfo.types import RegisterOffset

from .arch import Arch, Endness, Register, register_arch
from .archerror import ArchError
from .tls import TLSArchInfo

try:
    import capstone as _capstone
except ImportError:
    _capstone = None

try:
    import keystone as _keystone
except ImportError:
    _keystone = None

try:
    import unicorn as _unicorn
except ImportError:
    _unicorn = None

try:
    import pyvex as _pyvex
except ImportError:
    _pyvex = None


_NATIVE_FUNCTION_PROLOGS = {
    rb"\x55\x48\x89\xe5",  # push rbp; mov rbp, rsp
    rb"\x48[\x83,\x81]\xec[\x00-\xff]",  # sub rsp, xxx
}
# every function prolog can potentially be prefixed with endbr64
_endbr64 = b"\xf3\x0f\x1e\xfa"
_prefixed = {(_endbr64 + prolog) for prolog in _NATIVE_FUNCTION_PROLOGS}
_FUNCTION_PROLOGS = _prefixed | _NATIVE_FUNCTION_PROLOGS


class ArchAMD64(Arch):
    def __init__(self, endness=Endness.LE):
        if endness != Endness.LE:
            raise ArchError("Arch AMD64 must be little endian")
        super().__init__(endness)
        self.argument_register_positions = (
            {
                self.registers["rdi"][0]: 0,
                self.registers["rsi"][0]: 1,
                self.registers["rdx"][0]: 2,
                self.registers["rcx"][0]: 3,  # Used for user calls
                self.registers["r10"][0]: 3,  # Used for Linux kernel calls
                self.registers["r8"][0]: 4,
                self.registers["r9"][0]: 5,
                # fp registers
                self.registers["xmm0"][0]: 0,
                self.registers["xmm1"][0]: 1,
                self.registers["xmm2"][0]: 2,
                self.registers["xmm3"][0]: 3,
                self.registers["xmm4"][0]: 4,
                self.registers["xmm5"][0]: 5,
                self.registers["xmm6"][0]: 6,
                self.registers["xmm7"][0]: 7,
            }
            if _pyvex is not None
            else None
        )

        # Register blacklist
        reg_blacklist = ("fs", "gs")
        if self.reg_blacklist is not None and self.reg_blacklist_offsets is not None:
            for register in self.register_list:
                if register.name in reg_blacklist:
                    self.reg_blacklist.append(register.name)
                    self.reg_blacklist_offsets.append(register.vex_offset)

        if _unicorn and _pyvex:
            # CPU flag registers
            uc_flags_reg = _unicorn.x86_const.UC_X86_REG_EFLAGS
            cpu_flag_registers = {"d": 1 << 10, "ac": 1 << 18, "id": 1 << 21}
            for reg, reg_bitmask in cpu_flag_registers.items():
                reg_offset = self.get_register_offset(reg)
                self.cpu_flag_register_offsets_and_bitmasks_map[reg_offset] = (uc_flags_reg, reg_bitmask)

            mxcsr_registers = {"sseround": 1 << 14 | 1 << 13}
            uc_mxcsr_reg = _unicorn.x86_const.UC_X86_REG_MXCSR
            for reg, reg_bitmask in mxcsr_registers.items():
                reg_offset = self.get_register_offset(reg)
                self.cpu_flag_register_offsets_and_bitmasks_map[reg_offset] = (uc_mxcsr_reg, reg_bitmask)

    @property
    def capstone_x86_syntax(self):
        """
        The current syntax Capstone uses for x64. It can be 'intel' or 'at&t'
        """
        return self._cs_x86_syntax

    @capstone_x86_syntax.setter
    def capstone_x86_syntax(self, new_syntax):
        if new_syntax not in ("intel", "at&t"):
            raise ArchError('Unsupported Capstone x86 syntax. It must be either "intel" or "at&t".')

        if new_syntax != self._cs_x86_syntax:
            self._cs = None
            self._cs_x86_syntax = new_syntax

    def _configure_capstone(self):
        if self._cs_x86_syntax == "at&t":
            self._cs.syntax = _capstone.CS_OPT_SYNTAX_ATT
        else:
            self._cs.syntax = _capstone.CS_OPT_SYNTAX_INTEL

    @property
    def keystone_x86_syntax(self):
        """
        The current syntax Keystone uses for x86. It can be 'intel',
        'at&t', 'nasm', 'masm', 'gas' or 'radix16'
        """
        return self._ks_x86_syntax

    @keystone_x86_syntax.setter
    def keystone_x86_syntax(self, new_syntax):
        if new_syntax not in ("intel", "at&t", "nasm", "masm", "gas", "radix16"):
            raise ArchError(
                "Unsupported Keystone x86 syntax. It must be one of the following: "
                '"intel", "at&t", "nasm", "masm", "gas" or "radix16".'
            )

        if new_syntax != self._ks_x86_syntax:
            self._ks = None
            self._ks_x86_syntax = new_syntax

    def _configure_keystone(self):
        if self._ks_x86_syntax == "at&t":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_ATT
        elif self._ks_x86_syntax == "nasm":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_NASM
        elif self._ks_x86_syntax == "masm":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_MASM
        elif self._ks_x86_syntax == "gas":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_GAS
        elif self._ks_x86_syntax == "radix16":
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_RADIX16
        else:
            self._ks.syntax = _keystone.KS_OPT_SYNTAX_INTEL

    bits = 64
    vex_arch = "VexArchAMD64"
    vex_endness = "VexEndnessLE"
    name = "AMD64"
    pcode_id = "x86:LE:64:default"
    qemu_name = "x86_64"
    ida_processor = "metapc"
    linux_name = "x86_64"
    triplet = "x86_64-linux-gnu"
    max_inst_bytes = 15
    ret_offset = RegisterOffset(16)
    fp_ret_offset = RegisterOffset(224)  # xmm0
    vex_conditional_helpers = True
    syscall_num_offset = 16
    call_pushes_ret = True
    stack_change = -8
    initial_sp = 0x7FFFFFFF0000
    call_sp_fix = -8
    memory_endness = Endness.LE
    register_endness = Endness.LE
    sizeof = {"short": 16, "int": 32, "long": 64, "long long": 64}
    if _capstone:
        cs_arch = _capstone.CS_ARCH_X86
        cs_mode = _capstone.CS_MODE_64 + _capstone.CS_MODE_LITTLE_ENDIAN
    _cs_x86_syntax = None  # Set it to 'att' in order to use AT&T syntax for x86
    if _keystone:
        ks_arch = _keystone.KS_ARCH_X86
        ks_mode = _keystone.KS_MODE_64 + _keystone.KS_MODE_LITTLE_ENDIAN
    _ks_x86_syntax = None
    uc_arch = _unicorn.UC_ARCH_X86 if _unicorn else None
    uc_mode = (_unicorn.UC_MODE_64 + _unicorn.UC_MODE_LITTLE_ENDIAN) if _unicorn else None
    uc_const = _unicorn.x86_const if _unicorn else None
    uc_prefix = "UC_X86_" if _unicorn else None
    function_prologs = _FUNCTION_PROLOGS
    function_epilogs = {
        rb"\xc9\xc3",  # leaveq; retq
        rb"([^\x41][\x50-\x5f]{1}|\x41[\x50-\x5f])\xc3",  # pop <reg>; retq
        rb"\x48[\x83,\x81]\xc4([\x00-\xff]{1}|[\x00-\xff]{4})\xc3",  #  add rsp, <siz>; retq
    }
    ret_instruction = b"\xc3"
    nop_instruction = b"\x90"
    instruction_alignment = 1
    register_list = [
        Register(
            name="rax",
            size=8,
            subregisters=[("eax", 0, 4), ("ax", 0, 2), ("al", 0, 1), ("ah", 1, 1)],
            general_purpose=True,
            linux_entry_value=0x1C,
        ),
        Register(
            name="rcx",
            size=8,
            subregisters=[("ecx", 0, 4), ("cx", 0, 2), ("cl", 0, 1), ("ch", 1, 1)],
            general_purpose=True,
            argument=True,
        ),
        Register(
            name="rdx",
            size=8,
            subregisters=[("edx", 0, 4), ("dx", 0, 2), ("dl", 0, 1), ("dh", 1, 1)],
            general_purpose=True,
            argument=True,
            linux_entry_value="ld_destructor",
        ),
        Register(
            name="rbx",
            size=8,
            subregisters=[("ebx", 0, 4), ("bx", 0, 2), ("bl", 0, 1), ("bh", 1, 1)],
            general_purpose=True,
            linux_entry_value=0,
        ),
        Register(
            name="rsp",
            size=8,
            subregisters=[("esp", 0, 4)],
            alias_names=("sp",),
            general_purpose=True,
            default_value=(initial_sp, True, "global"),
        ),
        Register(
            name="rbp",
            size=8,
            subregisters=[("ebp", 0, 4), ("_bp", 0, 2), ("bpl", 0, 1), ("bph", 1, 1)],
            alias_names=("bp",),
            general_purpose=True,
            linux_entry_value=0,
        ),
        Register(
            name="rsi",
            size=8,
            subregisters=[("esi", 0, 4), ("si", 0, 2), ("sil", 0, 1), ("sih", 1, 1)],
            general_purpose=True,
            argument=True,
            linux_entry_value="argv",
        ),
        Register(
            name="rdi",
            size=8,
            subregisters=[("edi", 0, 4), ("di", 0, 2), ("dil", 0, 1), ("dih", 1, 1)],
            general_purpose=True,
            argument=True,
            linux_entry_value="argc",
        ),
        Register(
            name="r8",
            size=8,
            subregisters=[("r8d", 0, 4), ("r8w", 0, 2), ("r8b", 0, 1)],
            general_purpose=True,
            argument=True,
        ),
        Register(
            name="r9",
            size=8,
            subregisters=[("r9d", 0, 4), ("r9w", 0, 2), ("r9b", 0, 1)],
            general_purpose=True,
            argument=True,
        ),
        Register(
            name="r10",
            size=8,
            subregisters=[("r10d", 0, 4), ("r10w", 0, 2), ("r10b", 0, 1)],
            general_purpose=True,
            argument=True,
        ),
        Register(
            name="r11", size=8, subregisters=[("r11d", 0, 4), ("r11w", 0, 2), ("r11b", 0, 1)], general_purpose=True
        ),
        Register(
            name="r12",
            size=8,
            subregisters=[("r12d", 0, 4), ("r12w", 0, 2), ("r12b", 0, 1)],
            general_purpose=True,
            linux_entry_value=0,
        ),
        Register(
            name="r13",
            size=8,
            subregisters=[("r13d", 0, 4), ("r13w", 0, 2), ("r13b", 0, 1)],
            general_purpose=True,
            linux_entry_value=0,
        ),
        Register(
            name="r14",
            size=8,
            subregisters=[("r14d", 0, 4), ("r14w", 0, 2), ("r14b", 0, 1)],
            general_purpose=True,
            linux_entry_value=0,
        ),
        Register(
            name="r15",
            size=8,
            subregisters=[("r15d", 0, 4), ("r15w", 0, 2), ("r15b", 0, 1)],
            general_purpose=True,
            linux_entry_value=0,
        ),
        Register(name="cc_op", size=8, default_value=(0, False, None), concrete=False, artificial=True),
        Register(name="cc_dep1", size=8, concrete=False, artificial=True),
        Register(name="cc_dep2", size=8, concrete=False, artificial=True),
        Register(name="cc_ndep", size=8, concrete=False, artificial=True, linux_entry_value=0),
        Register(name="d", size=8, alias_names=("dflag",), default_value=(1, False, None), concrete=False),
        Register(name="rip", size=8, alias_names=("ip", "pc"), general_purpose=True),
        Register(name="ac", size=8, alias_names=("acflag",), concrete=False),
        Register(name="id", size=8, alias_names=("idflag",)),
        Register(
            name="fs",
            size=8,
            vex_name="fs_const",
            alias_names=("fs_const",),
            default_value=(0x9000000000000000, True, "global"),
            concrete=False,
        ),
        Register(name="sseround", size=8, vector=True, default_value=(0, False, None)),
        Register(name="cr0", size=8),
        Register(name="cr2", size=8),
        Register(name="cr3", size=8),
        Register(name="cr4", size=8),
        Register(name="cr8", size=8),
        Register(
            name="ymm0",
            size=32,
            subregisters=[("xmm0", 0, 16), ("xmm0lq", 0, 8), ("xmm0hq", 8, 8), ("ymm0hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm1",
            size=32,
            subregisters=[("xmm1", 0, 16), ("xmm1lq", 0, 8), ("xmm1hq", 8, 8), ("ymm1hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm2",
            size=32,
            subregisters=[("xmm2", 0, 16), ("xmm2lq", 0, 8), ("xmm2hq", 8, 8), ("ymm2hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm3",
            size=32,
            subregisters=[("xmm3", 0, 16), ("xmm3lq", 0, 8), ("xmm3hq", 8, 8), ("ymm3hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm4",
            size=32,
            subregisters=[("xmm4", 0, 16), ("xmm4lq", 0, 8), ("xmm4hq", 8, 8), ("ymm4hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm5",
            size=32,
            subregisters=[("xmm5", 0, 16), ("xmm5lq", 0, 8), ("xmm5hq", 8, 8), ("ymm5hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm6",
            size=32,
            subregisters=[("xmm6", 0, 16), ("xmm6lq", 0, 8), ("xmm6hq", 8, 8), ("ymm6hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm7",
            size=32,
            subregisters=[("xmm7", 0, 16), ("xmm7lq", 0, 8), ("xmm7hq", 8, 8), ("ymm7hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm8",
            size=32,
            subregisters=[("xmm8", 0, 16), ("xmm8lq", 0, 8), ("xmm8hq", 8, 8), ("ymm8hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm9",
            size=32,
            subregisters=[("xmm9", 0, 16), ("xmm9lq", 0, 8), ("xmm9hq", 8, 8), ("ymm9hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm10",
            size=32,
            subregisters=[("xmm10", 0, 16), ("xmm10lq", 0, 8), ("xmm10hq", 8, 8), ("ymm10hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm11",
            size=32,
            subregisters=[("xmm11", 0, 16), ("xmm11lq", 0, 8), ("xmm11hq", 8, 8), ("ymm11hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm12",
            size=32,
            subregisters=[("xmm12", 0, 16), ("xmm12lq", 0, 8), ("xmm12hq", 8, 8), ("ymm12hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm13",
            size=32,
            subregisters=[("xmm13", 0, 16), ("xmm13lq", 0, 8), ("xmm13hq", 8, 8), ("ymm13hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm14",
            size=32,
            subregisters=[("xmm14", 0, 16), ("xmm14lq", 0, 8), ("xmm14hq", 8, 8), ("ymm14hx", 16, 16)],
            vector=True,
        ),
        Register(
            name="ymm15",
            size=32,
            subregisters=[("xmm15", 0, 16), ("xmm15lq", 0, 8), ("xmm15hq", 8, 8), ("ymm15hx", 16, 16)],
            vector=True,
        ),
        Register(name="ftop", size=4, floating_point=True, default_value=(7, False, None), artificial=True),
        Register(
            name="fpreg",
            size=64,
            subregisters=[
                ("mm0", 0, 8),
                ("mm1", 8, 8),
                ("mm2", 16, 8),
                ("mm3", 24, 8),
                ("mm4", 32, 8),
                ("mm5", 40, 8),
                ("mm6", 48, 8),
                ("mm7", 56, 8),
            ],
            alias_names=("fpu_regs",),
            floating_point=True,
        ),
        Register(name="fptag", size=8, alias_names=("fpu_tags",), floating_point=True, default_value=(0, False, None)),
        Register(name="fpround", size=8, floating_point=True, default_value=(0, False, None)),
        Register(name="fc3210", size=8, floating_point=True),
        Register(name="emnote", size=4, artificial=True),
        Register(name="cmstart", size=8),
        Register(name="cmlen", size=8),
        Register(name="nraddr", size=8),
        Register(name="gs", size=8, vex_name="gs_const", alias_names=("gs_const",), concrete=False),
        Register(name="ip_at_syscall", size=8, concrete=False, artificial=True),
        Register(name="cs_seg", size=2, vex_name="cs"),
        Register(name="ds_seg", size=2, vex_name="ds"),
        Register(name="es_seg", size=2, vex_name="es"),
        Register(name="fs_seg", size=2, vex_name="fs"),
        Register(name="gs_seg", size=2, vex_name="gs"),
        Register(name="ss_seg", size=2, vex_name="ss"),
    ]

    # https://gitlab.com/x86-psABIs/x86-64-ABI
    dynamic_tag_translation = {
        0x70000000: "DT_X86_64_PLT",
        0x70000001: "DT_X86_64_PLTSZ",
        0x70000003: "DT_X86_64_PLTENT",
    }

    symbol_type_translation = {10: "STT_GNU_IFUNC", "STT_LOOS": "STT_GNU_IFUNC"}
    got_section_name = ".got.plt"
    ld_linux_name = "ld-linux-x86-64.so.2"
    elf_tls = TLSArchInfo(2, 704, [16], [8], [0], 0, 0)
    dwarf_registers = [
        "rax",
        "rdx",
        "rcx",
        "rbx",
        "rsi",
        "rdi",
        "rbp",
        "rsp",
        "r8",
        "r9",
        "r10",
        "r11",
        "r12",
        "r13",
        "r14",
        "r15",
        "rip",
        "xmm0",
        "xmm1",
        "xmm2",
        "xmm3",
        "xmm4",
        "xmm5",
        "xmm6",
        "xmm7",
        "xmm8",
        "xmm9",
        "xmm10",
        "xmm11",
        "xmm12",
        "xmm13",
        "xmm14",
        "xmm15",
        "st0",
        "st1",
        "st2",
        "st3",
        "st4",
        "st5",
        "st6",
        "st7",
        "mm0",
        "mm1",
        "mm2",
        "mm3",
        "mm4",
        "mm5",
        "mm6",
        "mm7",
        "rflags",
        "es",
        "cs",
        "ss",
        "ds",
        "fs",
        "gs",
        "<none>",
        "<none>",
        "fs.base",
        "gs.base",
        "<none>",
        "<none>",
        "tr",
        "ldtr",
        "mxcsr",
        "fcw",
        "fsw",
    ]


register_arch([r".*amd64|.*x64|.*x86_64|.*metapc"], 64, Endness.LE, ArchAMD64)



from archinfo.types import RegisterOffset

from .arch import Arch, Endness, Register, register_arch
from .tls import TLSArchInfo

try:
    import capstone as _capstone
except ImportError:
    _capstone = None

try:
    import keystone as _keystone
except ImportError:
    _keystone = None

try:
    import unicorn as _unicorn
except ImportError:
    _unicorn = None


class ArchMIPS64(Arch):
    def __init__(self, endness=Endness.BE):
        super().__init__(endness)
        if endness == Endness.BE:
            self.pcode_id = "MIPS:BE:64:default"
            self.function_prologs = set()
            self.function_epilogs = set()
            self.triplet = "mips64-linux-gnu"
            self.linux_name = "mips64"
            self.ida_name = "mips64b"
        else:
            self.pcode_id = "MIPS:LE:64:default"

    bits = 64
    vex_arch = "VexArchMIPS64"
    name = "MIPS64"
    default_endness = Endness.BE
    qemu_name = "mips64el"
    ida_processor = "mips64"
    linux_name = "mips64el"  # ???
    triplet = "mips64el-linux-gnu"
    max_inst_bytes = 4
    ret_offset = RegisterOffset(32)
    syscall_register_offset = 16
    call_pushes_ret = False
    stack_change = -8
    branch_delay_slot = True
    sizeof = {"short": 16, "int": 32, "long": 64, "long long": 64}
    if _capstone:
        cs_arch = _capstone.CS_ARCH_MIPS
        cs_mode = _capstone.CS_MODE_64 + _capstone.CS_MODE_LITTLE_ENDIAN
    if _keystone:
        ks_arch = _keystone.KS_ARCH_MIPS
        ks_mode = _keystone.KS_MODE_64 + _keystone.KS_MODE_LITTLE_ENDIAN
    uc_arch = _unicorn.UC_ARCH_MIPS if _unicorn else None
    uc_mode = (_unicorn.UC_MODE_64 + _unicorn.UC_MODE_LITTLE_ENDIAN) if _unicorn else None
    uc_const = _unicorn.mips_const if _unicorn else None
    uc_prefix = "UC_MIPS_" if _unicorn else None
    function_prologs = set()
    function_epilogs = set()

    ret_instruction = b"\x08\x00\xe0\x03" + b"\x25\x08\x20\x00"
    nop_instruction = b"\x00\x00\x00\x00"
    instruction_alignment = 4
    register_list = [
        Register(name="zero", size=8, alias_names=("r0",)),
        Register(name="at", size=8, alias_names=("r1",), general_purpose=True),
        Register(name="v0", size=8, alias_names=("r2",), general_purpose=True, linux_entry_value="ld_destructor"),
        Register(name="v1", size=8, alias_names=("r3",), general_purpose=True),
        Register(name="a0", size=8, alias_names=("r4",), general_purpose=True, argument=True),
        Register(name="a1", size=8, alias_names=("r5",), general_purpose=True, argument=True),
        Register(name="a2", size=8, alias_names=("r6",), general_purpose=True, argument=True),
        Register(name="a3", size=8, alias_names=("r7",), general_purpose=True, argument=True),
        Register(
            name="t0",
            size=8,
            alias_names=(
                "r8",
                "a4",
            ),
            general_purpose=True,
        ),
        Register(
            name="t1",
            size=8,
            alias_names=(
                "r9",
                "a5",
            ),
            general_purpose=True,
        ),
        Register(
            name="t2",
            size=8,
            alias_names=(
                "r10",
                "a6",
            ),
            general_purpose=True,
        ),
        Register(
            name="t3",
            size=8,
            alias_names=(
                "r11",
                "a7",
            ),
            general_purpose=True,
        ),
        Register(name="t4", size=8, alias_names=("r12",), general_purpose=True),
        Register(name="t5", size=8, alias_names=("r13",), general_purpose=True),
        Register(name="t6", size=8, alias_names=("r14",), general_purpose=True),
        Register(name="t7", size=8, alias_names=("r15",), general_purpose=True),
        Register(name="s0", size=8, alias_names=("r16",), general_purpose=True),
        Register(name="s1", size=8, alias_names=("r17",), general_purpose=True),
        Register(name="s2", size=8, alias_names=("r18",), general_purpose=True),
        Register(name="s3", size=8, alias_names=("r19",), general_purpose=True),
        Register(name="s4", size=8, alias_names=("r20",), general_purpose=True),
        Register(name="s5", size=8, alias_names=("r21",), general_purpose=True),
        Register(name="s6", size=8, alias_names=("r22",), general_purpose=True),
        Register(name="s7", size=8, alias_names=("r23",), general_purpose=True),
        Register(name="t8", size=8, alias_names=("r24",), general_purpose=True),
        Register(name="t9", size=8, alias_names=("r25",), general_purpose=True, persistent=True),
        Register(name="k0", size=8, alias_names=("r26",), general_purpose=True),
        Register(name="k1", size=8, alias_names=("r27",), general_purpose=True),
        Register(name="gp", size=8, alias_names=("r28",), persistent=True),
        Register(name="sp", size=8, alias_names=("r29",), default_value=(Arch.initial_sp, True, "global")),
        Register(name="s8", size=8, alias_names=("r30", "fp", "bp"), general_purpose=True),
        Register(
            name="ra", size=8, alias_names=("r31", "lr"), general_purpose=True, persistent=True, linux_entry_value=0
        ),
        Register(name="pc", size=8, alias_names=("ip",)),
        Register(name="hi", size=8, general_purpose=True),
        Register(name="lo", size=8, general_purpose=True),
        Register(name="f0", size=8, floating_point=True, subregisters=[("f0_lo", 0, 4)]),
        Register(name="f1", size=8, floating_point=True, subregisters=[("f1_lo", 0, 4)]),
        Register(name="f2", size=8, floating_point=True, subregisters=[("f2_lo", 0, 4)]),
        Register(name="f3", size=8, floating_point=True, subregisters=[("f3_lo", 0, 4)]),
        Register(name="f4", size=8, floating_point=True, subregisters=[("f4_lo", 0, 4)]),
        Register(name="f5", size=8, floating_point=True, subregisters=[("f5_lo", 0, 4)]),
        Register(name="f6", size=8, floating_point=True, subregisters=[("f6_lo", 0, 4)]),
        Register(name="f7", size=8, floating_point=True, subregisters=[("f7_lo", 0, 4)]),
        Register(name="f8", size=8, floating_point=True, subregisters=[("f8_lo", 0, 4)]),
        Register(name="f9", size=8, floating_point=True, subregisters=[("f9_lo", 0, 4)]),
        Register(name="f10", size=8, floating_point=True, subregisters=[("f10_lo", 0, 4)]),
        Register(name="f11", size=8, floating_point=True, subregisters=[("f11_lo", 0, 4)]),
        Register(name="f12", size=8, floating_point=True, subregisters=[("f12_lo", 0, 4)]),
        Register(name="f13", size=8, floating_point=True, subregisters=[("f13_lo", 0, 4)]),
        Register(name="f14", size=8, floating_point=True, subregisters=[("f14_lo", 0, 4)]),
        Register(name="f15", size=8, floating_point=True, subregisters=[("f15_lo", 0, 4)]),
        Register(name="f16", size=8, floating_point=True, subregisters=[("f16_lo", 0, 4)]),
        Register(name="f17", size=8, floating_point=True, subregisters=[("f17_lo", 0, 4)]),
        Register(name="f18", size=8, floating_point=True, subregisters=[("f18_lo", 0, 4)]),
        Register(name="f19", size=8, floating_point=True, subregisters=[("f19_lo", 0, 4)]),
        Register(name="f20", size=8, floating_point=True, subregisters=[("f20_lo", 0, 4)]),
        Register(name="f21", size=8, floating_point=True, subregisters=[("f21_lo", 0, 4)]),
        Register(name="f22", size=8, floating_point=True, subregisters=[("f22_lo", 0, 4)]),
        Register(name="f23", size=8, floating_point=True, subregisters=[("f23_lo", 0, 4)]),
        Register(name="f24", size=8, floating_point=True, subregisters=[("f24_lo", 0, 4)]),
        Register(name="f25", size=8, floating_point=True, subregisters=[("f25_lo", 0, 4)]),
        Register(name="f26", size=8, floating_point=True, subregisters=[("f26_lo", 0, 4)]),
        Register(name="f27", size=8, floating_point=True, subregisters=[("f27_lo", 0, 4)]),
        Register(name="f28", size=8, floating_point=True, subregisters=[("f28_lo", 0, 4)]),
        Register(name="f29", size=8, floating_point=True, subregisters=[("f29_lo", 0, 4)]),
        Register(name="f30", size=8, floating_point=True, subregisters=[("f30_lo", 0, 4)]),
        Register(name="f31", size=8, floating_point=True, subregisters=[("f31_lo", 0, 4)]),
        Register(name="fir", size=4, floating_point=True),
        Register(name="fccr", size=4, floating_point=True),
        Register(name="fexr", size=4, floating_point=True),
        Register(name="fenr", size=4, floating_point=True),
        Register(name="fcsr", size=4, floating_point=True),
        Register(name="cp0_status", size=4),
        Register(name="ulr", size=8),
        Register(name="emnote", size=4, artificial=True),
        Register(name="cond", size=4),
        Register(name="cmstart", size=8),
        Register(name="cmlen", size=8),
        Register(name="nraddr", size=8),
        Register(name="ip_at_syscall", size=8, artificial=True),
    ]

    # http://techpubs.sgi.com/library/manuals/4000/007-4658-001/pdf/007-4658-001.pdf
    dynamic_tag_translation = {
        0x70000001: "DT_MIPS_RLD_VERSION",
        0x70000005: "DT_MIPS_FLAGS",
        0x70000006: "DT_MIPS_BASE_ADDRESS",
        0x7000000A: "DT_MIPS_LOCAL_GOTNO",
        0x70000011: "DT_MIPS_SYMTABNO",
        0x70000012: "DT_MIPS_UNREFEXTNO",
        0x70000013: "DT_MIPS_GOTSYM",
        0x70000016: "DT_MIPS_RLD_MAP",
    }
    got_section_name = ".got"
    ld_linux_name = "ld.so.1"
    elf_tls = TLSArchInfo(1, 16, [], [0], [], 0x7000, 0x8000)


register_arch([r".*mipsel.*|.*mips64el|.*mipsel64"], 64, Endness.LE, ArchMIPS64)
register_arch([r".*mips64.*|.*mips.*"], 64, Endness.ANY, ArchMIPS64)



from archinfo.types import RegisterOffset

from .arch import Arch, Endness, Register, register_arch
from .tls import TLSArchInfo

try:
    import capstone as _capstone
except ImportError:
    _capstone = None

try:
    import keystone as _keystone
except ImportError:
    _keystone = None

try:
    import unicorn as _unicorn
    from unicorn import riscv_const
except ImportError:
    _unicorn = None
    riscv_const = None


def is_riscv_arch(a):
    return a.name.startswith("RISCV")


class ArchRISCV64(Arch):
    def __init__(self, endness=Endness.LE):
        super().__init__(endness)
        if endness == Endness.BE:
            self.function_prologs = set()
            self.function_epilogs = set()

    bits = 64
    vex_arch = "VexArchRISCV64"
    name = "RISCV64"
    pcode_id = "RISCV:LE:64:default"
    qemu_name = "riscv64"
    linux_name = "riscv64"
    triplet = "riscv64-linux-gnu"
    max_inst_bytes = 4
    ret_offset = RegisterOffset(4)
    vex_conditional_helpers = True
    syscall_num_offset = 152
    call_pushes_ret = False
    stack_change = -8
    initial_sp = 0x7FFFFFFFFFF0000
    memory_endness = Endness.LE
    register_endness = Endness.LE
    instruction_endness = Endness.LE
    sizeof = {"short": 16, "int": 32, "long": 64, "long long": 64}
    if _capstone and hasattr(_capstone, "CS_ARCH_RISCV"):
        cs_arch = _capstone.CS_ARCH_RISCV
        cs_mode = _capstone.CS_MODE_RISCV64 | _capstone.CS_MODE_RISCVC
    # if _keystone:
    #     ks_arch = _keystone.KS_ARCH_RISCV
    #     ks_mode = _keystone.KS_MODE_LITTLE_ENDIAN
    uc_arch = _unicorn.UC_ARCH_RISCV if _unicorn else None
    uc_mode = _unicorn.UC_MODE_RISCV64 if _unicorn else None
    uc_const = riscv_const if _unicorn else None
    uc_prefix = "UC_RISCV_" if _unicorn else None

    ret_instruction = b"\x82\x80"  # ret
    nop_instruction = b"\x01\x00"  # nop
    function_prologs = set()
    function_epilogs = set()
    instruction_alignment = 2
    register_list = [
        Register(name="zero", size=8, alias_names=("x0",)),
        Register(
            name="ra",
            size=8,
            alias_names=(
                "x1",
                "lr",
            ),
            general_purpose=True,
        ),
        Register(name="sp", size=8, alias_names=("x2",), general_purpose=True),
        Register(name="gp", size=8, alias_names=("x3",), general_purpose=True),
        Register(name="tp", size=8, alias_names=("x4",), general_purpose=True),
        Register(name="t0", size=8, alias_names=("x5",), general_purpose=True),
        Register(name="t1", size=8, alias_names=("x6",), general_purpose=True),
        Register(name="t2", size=8, alias_names=("x7",), general_purpose=True),
        Register(name="s0", size=8, alias_names=("bp", "fp", "x8"), general_purpose=True, vex_offset=80),
        Register(name="s1", size=8, alias_names=("x9",), general_purpose=True),
        Register(name="a0", size=8, alias_names=("x10",), general_purpose=True, argument=True),
        Register(name="a1", size=8, alias_names=("x11",), general_purpose=True, argument=True),
        Register(name="a2", size=8, alias_names=("x12",), general_purpose=True, argument=True),
        Register(name="a3", size=8, alias_names=("x13",), general_purpose=True, argument=True),
        Register(name="a4", size=8, alias_names=("x14",), general_purpose=True, argument=True),
        Register(name="a5", size=8, alias_names=("x15",), general_purpose=True, argument=True),
        Register(name="a6", size=8, alias_names=("x16",), general_purpose=True, argument=True),
        Register(name="a7", size=8, alias_names=("x17",), general_purpose=True, argument=True),
        Register(name="s2", size=8, alias_names=("x18",), general_purpose=True),
        Register(name="s3", size=8, alias_names=("x19",), general_purpose=True),
        Register(name="s4", size=8, alias_names=("x20",), general_purpose=True),
        Register(name="s5", size=8, alias_names=("x21",), general_purpose=True),
        Register(name="s6", size=8, alias_names=("x22",), general_purpose=True),
        Register(name="s7", size=8, alias_names=("x23",), general_purpose=True),
        Register(name="s8", size=8, alias_names=("x24",), general_purpose=True),
        Register(name="s9", size=8, alias_names=("x25",), general_purpose=True),
        Register(name="s10", size=8, alias_names=("x26",), general_purpose=True),
        Register(name="s11", size=8, alias_names=("x27",), general_purpose=True),
        Register(name="t3", size=8, alias_names=("x28",), general_purpose=True),
        Register(name="t4", size=8, alias_names=("x29",), general_purpose=True),
        Register(name="t5", size=8, alias_names=("x30",), general_purpose=True),
        Register(name="t6", size=8, alias_names=("x31",), general_purpose=True),
        Register(name="pc", size=8, alias_names=("ip",)),
        Register(name="ft0", size=8, alias_names=("f0",), floating_point=True),
        Register(name="ft1", size=8, alias_names=("f1",), floating_point=True),
        Register(name="ft2", size=8, alias_names=("f2",), floating_point=True),
        Register(name="ft3", size=8, alias_names=("f3",), floating_point=True),
        Register(name="ft4", size=8, alias_names=("f4",), floating_point=True),
        Register(name="ft5", size=8, alias_names=("f5",), floating_point=True),
        Register(name="ft6", size=8, alias_names=("f6",), floating_point=True),
        Register(name="ft7", size=8, alias_names=("f7",), floating_point=True),
        Register(name="fs0", size=8, alias_names=("f8",), floating_point=True, vex_offset=344),
        Register(name="fs1", size=8, alias_names=("f9",), floating_point=True),
        Register(name="fa0", size=8, alias_names=("f10",), floating_point=True),
        Register(name="fa1", size=8, alias_names=("f11",), floating_point=True),
        Register(name="fa2", size=8, alias_names=("f12",), floating_point=True),
        Register(name="fa3", size=8, alias_names=("f13",), floating_point=True),
        Register(name="fa4", size=8, alias_names=("f14",), floating_point=True),
        Register(name="fa5", size=8, alias_names=("f15",), floating_point=True),
        Register(name="fa6", size=8, alias_names=("f16",), floating_point=True),
        Register(name="fa7", size=8, alias_names=("f17",), floating_point=True),
        Register(name="fs2", size=8, alias_names=("f18",), floating_point=True),
        Register(name="fs3", size=8, alias_names=("f19",), floating_point=True),
        Register(name="fs4", size=8, alias_names=("f20",), floating_point=True),
        Register(name="fs5", size=8, alias_names=("f21",), floating_point=True),
        Register(name="fs6", size=8, alias_names=("f22",), floating_point=True),
        Register(name="fs7", size=8, alias_names=("f23",), floating_point=True),
        Register(name="fs8", size=8, alias_names=("f24",), floating_point=True),
        Register(name="fs9", size=8, alias_names=("f25",), floating_point=True),
        Register(name="fs10", size=8, alias_names=("f26",), floating_point=True),
        Register(name="fs11", size=8, alias_names=("f27",), floating_point=True),
        Register(name="ft8", size=8, alias_names=("f28",), floating_point=True),
        Register(name="ft9", size=8, alias_names=("f29",), floating_point=True),
        Register(name="ft10", size=8, alias_names=("f30",), floating_point=True),
        Register(name="ft11", size=8, alias_names=("f31",), floating_point=True),
        Register(name="ip_at_syscall", size=8, artificial=True),
    ]

    got_section_name = ".got"
    ld_linux_name = "ld-linux-riscv64-lp64d.so.1"
    elf_tls = TLSArchInfo(1, 32, [], [0], [], 0, 0)


register_arch([r".*riscv.*"], 64, Endness.ANY, ArchRISCV64)



import logging
from typing import Union

from .arch import Arch, Endness, Register
from .archerror import ArchError
from .tls import TLSArchInfo
from .types import RegisterOffset

try:
    import pypcode

    _has_pypcode = True
except ImportError:
    _has_pypcode = False


log = logging.getLogger(__name__)


class ArchPcode(Arch):
    """
    archinfo interface to pypcode architectures. Provides minimal mapping for
    architectural info like register file map, endianness, bit width, etc.
    """

    def __init__(self, language: Union["pypcode.ArchLanguage", str]):
        if not _has_pypcode:
            raise ArchError("pypcode not installed")

        if isinstance(language, str):
            language = self._get_language_by_id(language)

        assert isinstance(language, pypcode.ArchLanguage)

        self.name = language.id
        self.pcode_id = language.id
        self.description = language.description
        self.bits = int(language.size)
        self.endness = {"little": Endness.LE, "big": Endness.BE}[language.endian]
        self.instruction_endness = self.endness
        self.sizeof = (
            {"short": 16, "int": 32, "long": 64, "long long": 64}
            if self.bits == 64
            else {"short": 16, "int": 32, "long": 32, "long long": 64}
        )
        self.elf_tls = TLSArchInfo(1, 8, [], [0], [], 0, 0)

        # Build registers list
        ctx = pypcode.Context(language)
        archinfo_regs = {rname.lower(): Register(rname.lower(), r.size, r.offset) for rname, r in ctx.registers.items()}

        # Get program counter register
        pc_offset = None
        pc_tag = language.pspec.find("programcounter") if language.pspec is not None else None
        if pc_tag is not None:
            pc_reg = pc_tag.attrib.get("register", None)
            if pc_reg is not None:
                # FIXME: Assumes RAM space
                pc_offset = ctx.registers[pc_reg].offset
                aliases = {"pc", "ip"}
                aliases.discard(pc_reg.lower())
                for alias in aliases:
                    archinfo_regs.pop(alias, None)
                archinfo_regs[pc_reg.lower()].alias_names = tuple(aliases)

        if pc_offset is None:
            log.warning("Unknown program counter register offset?")
            pc_offset = 0x80000000

        sp_offset = None
        ret_offset = RegisterOffset(0)
        if len(language.cspecs):

            def find_matching_cid(language, desired):
                for cid in language.cspecs:
                    if cid[0] == desired:
                        return cid
                return None

            cspec_id = (
                find_matching_cid(language, "default") or find_matching_cid(language, "gcc") or list(language.cspecs)[0]
            )
            cspec = language.cspecs[cspec_id]

            # Get stack pointer register
            sp_tag = cspec.find("stackpointer")
            if sp_tag is not None:
                sp_reg = sp_tag.attrib.get("register", None)
                if sp_reg is not None:
                    # FIXME: Assumes RAM space
                    sp_offset = ctx.registers[sp_reg].offset
                    if sp_reg.lower() != "sp":
                        if "sp" in archinfo_regs:
                            log.warning("Unexpected SP conflict")
                            del archinfo_regs["sp"]
                        archinfo_regs[sp_reg.lower()].alias_names += ("sp",)

            # Get return offset
            proto_tags = cspec.find("default_proto")
            if proto_tags is not None and len(proto_tags) >= 1:
                proto_tag = proto_tags[0]
                output_tags = proto_tag.find("output")
                if output_tags is not None and len(output_tags) >= 1:
                    output_tag = output_tags[0]
                    output_register_tag = output_tag.find("register")
                    if output_register_tag is not None:
                        output_reg = output_register_tag.attrib["name"]
                        ret_offset = RegisterOffset(ctx.registers[output_reg].offset)

        if sp_offset is None:
            log.warning("Unknown stack pointer register offset?")
            sp_offset = 0x80000008

        self.instruction_alignment = 1
        self.ip_offset = RegisterOffset(pc_offset)
        self.sp_offset = RegisterOffset(sp_offset)
        self.bp_offset = RegisterOffset(sp_offset)
        self.ret_offset = RegisterOffset(ret_offset)
        self.register_list = list(archinfo_regs.values())
        self.initial_sp = (0x8000 << (self.bits - 16)) - 1
        self.linux_name = ""  # FIXME
        self.triplet = ""  # FIXME

        # TODO: Replace the following hardcoded function prologues by data sourced from patterns.xml
        if "PowerPC:BE" in self.name:
            self.function_prologs = {
                # stwu  r1, xx(r1); mfspr rx, lr
                b"\x94\x21[\xc0-\xff][\x00\x10\x20\x30\x40\x50\x60\x70\x80\x90\xa0\xb0\xc0\xd0\xe0\xf0]"
                b"[\x7c-\x7f][\x08\x28\x48\x68\x88\xa8\xc8\xe8]\x02\xa6",
            }
        if self.name.startswith("Xtensa:LE"):
            self.function_prologs = {
                # entry  a1, N
                b"\x36[\x11\x21\x31\x41\x51\x61\x71\x81\x91\xa1\xb1\xc1\xd1\xe1\xf1]\x00",
            }

        if "sparc:" in self.name.lower() and self.bits == 32:
            self.branch_delay_slot = True

        super().__init__(endness=self.endness, instruction_endness=self.instruction_endness)

    def pcode_arch(self) -> "ArchPcode":
        """
        Return the ArchPcode instance for this architecture.
        For ArchPcode, this returns itself.

        :return: This ArchPcode instance
        """
        return self

    @staticmethod
    def _get_language_by_id(lang_id) -> "pypcode.ArchLanguage":
        if not _has_pypcode:
            raise ArchError("pypcode not installed")
        for arch in pypcode.Arch.enumerate():
            for lang in arch.languages:
                if lang.id == lang_id:
                    return lang
        raise ArchError("Language not found")

    def disasm(self, bytestring, addr=0, thumb=False):
        ctx = pypcode.Context(self.name)
        ctx.setVariableDefault("TMode", 1 if thumb else 0)
        instructions = ctx.disassemble(bytestring, addr, 0, len(bytestring), 0).instructions
        return "\n".join(f"{insn.addr.offset:#x}:\t{insn.mnem} {insn.body}" for insn in instructions)
